@@ -5,7 +5,7 @@ Interrupt and signal handling for Cython
 /*****************************************************************************
  *       Copyright (C) 2006 William Stein <wstein@gmail.com>
  *                     2006-2016 Martin Albrecht <martinralbrecht+cysignals@gmail.com>
- *                     2010-2016 Jeroen Demeyer <jdemeyer@cage.ugent.be>
+ *                     2010-2018 Jeroen Demeyer <J.Demeyer@UGent.be>
  *
  * cysignals is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License as published
@@ -22,14 +22,23 @@ Interrupt and signal handling for Cython
  *
  ****************************************************************************/
 
+
+#if __USE_FORTIFY_LEVEL
+#error "cysignals must be compiled without _FORTIFY_SOURCE"
+#endif
+
+
 #include "config.h"
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 #include <limits.h>
+#include <errno.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <sys/mman.h>
+#include <pthread.h>
 #if HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -48,7 +57,11 @@ static int PARI_SIGINT_block = 0;
 static int PARI_SIGINT_pending = 0;
 #endif
 #include "struct_signals.h"
-#include "signals.h"
+
+/* Some systems have MAP_ANON but not MAP_ANONYMOUS */
+#ifndef MAP_ANONYMOUS
+#define MAP_ANONYMOUS MAP_ANON
+#endif
 
 
 #if ENABLE_DEBUG_CYSIGNALS
@@ -63,14 +76,23 @@ static cysigs_t cysigs;
  * initialized by setup_cysignals_handlers(). */
 static sigset_t default_sigmask;
 
+/* A trampoline to jump to after handling a signal. */
+static cyjmp_buf trampoline_setup;
+static sigjmp_buf trampoline;
+
 /* default_sigmask with SIGHUP, SIGINT, SIGALRM added. */
 static sigset_t sigmask_with_sigint;
 
 
 static void do_raise_exception(int sig);
 static void sigdie(int sig, const char* s);
+
+#define BACKTRACELEN 1024
 static void print_backtrace(void);
 
+/* Implemented in signals.pyx */
+static int sig_raise_exception(int sig, const char* msg);
+
 
 /* Do whatever is needed to reset the CPU to a sane state after
  * handling a signals.  In particular on x86 CPUs, we need to clear
@@ -92,6 +114,61 @@ static inline void reset_CPU(void)
 }
 
 
+/* Reset all signal handlers and the signal mask to their defaults. */
+static inline void sig_reset_defaults(void) {
+    signal(SIGHUP, SIG_DFL);
+    signal(SIGINT, SIG_DFL);
+    signal(SIGQUIT, SIG_DFL);
+    signal(SIGILL, SIG_DFL);
+    signal(SIGABRT, SIG_DFL);
+    signal(SIGFPE, SIG_DFL);
+    signal(SIGBUS, SIG_DFL);
+    signal(SIGSEGV, SIG_DFL);
+    signal(SIGALRM, SIG_DFL);
+    signal(SIGTERM, SIG_DFL);
+    sigprocmask(SIG_SETMASK, &default_sigmask, NULL);
+}
+
+
+/* Call sigdie() with the appropriate error message for the signal, or
+ * whether the exception occurred inside our signal handler */
+static inline void sigdie_for_sig(int sig, int inside)
+{
+    sig_reset_defaults();
+    if (inside) sigdie(sig, "An error occurred during signal handling.");
+
+    /* Quit Python with an appropriate message. */
+    switch(sig)
+    {
+        case SIGQUIT:
+            sigdie(sig, NULL);
+            break;  /* This will not be reached */
+        case SIGILL:
+            sigdie(sig, "Unhandled SIGILL: An illegal instruction occurred.");
+            break;  /* This will not be reached */
+        case SIGABRT:
+            sigdie(sig, "Unhandled SIGABRT: An abort() occurred.");
+            break;  /* This will not be reached */
+        case SIGFPE:
+            sigdie(sig, "Unhandled SIGFPE: An unhandled floating point exception occurred.");
+            break;  /* This will not be reached */
+        case SIGBUS:
+            sigdie(sig, "Unhandled SIGBUS: A bus error occurred.");
+            break;  /* This will not be reached */
+        case SIGSEGV:
+            sigdie(sig, "Unhandled SIGSEGV: A segmentation fault occurred.");
+            break;  /* This will not be reached */
+    };
+    sigdie(sig, "Unknown signal received.\n");
+}
+
+
+/* Additional platform-specific implementation code */
+#if defined(__CYGWIN__)
+#include "implementation_cygwin.c"
+#endif
+
+
 /* Handler for SIGHUP, SIGINT, SIGALRM
  *
  * Inside sig_on() (i.e. when cysigs.sig_on_count is positive), this
@@ -119,8 +196,7 @@ static void cysigs_interrupt_handler(int sig)
             do_raise_exception(sig);
 
             /* Jump back to sig_on() (the first one if there is a stack) */
-            reset_CPU();
-            siglongjmp(cysigs.env, sig);
+            siglongjmp(trampoline, sig);
         }
     }
     else
@@ -167,8 +243,7 @@ static void cysigs_signal_handler(int sig)
         do_raise_exception(sig);
 
         /* Jump back to sig_on() (the first one if there is a stack) */
-        reset_CPU();
-        siglongjmp(cysigs.env, sig);
+        siglongjmp(trampoline, sig);
     }
     else
     {
@@ -176,48 +251,87 @@ static void cysigs_signal_handler(int sig)
 
         /* Reset all signals to their default behaviour and unblock
          * them in case something goes wrong as of now. */
-        signal(SIGHUP, SIG_DFL);
-        signal(SIGINT, SIG_DFL);
-        signal(SIGQUIT, SIG_DFL);
-        signal(SIGILL, SIG_DFL);
-        signal(SIGABRT, SIG_DFL);
-        signal(SIGFPE, SIG_DFL);
-        signal(SIGBUS, SIG_DFL);
-        signal(SIGSEGV, SIG_DFL);
-        signal(SIGALRM, SIG_DFL);
-        signal(SIGTERM, SIG_DFL);
-        sigprocmask(SIG_SETMASK, &default_sigmask, NULL);
-
-        if (inside) sigdie(sig, "An error occurred during signal handling.");
-
-        /* Quit Python with an appropriate message. */
-        switch(sig)
-        {
-            case SIGQUIT:
-                sigdie(sig, NULL);
-                break;  /* This will not be reached */
-            case SIGILL:
-                sigdie(sig, "Unhandled SIGILL: An illegal instruction occurred.");
-                break;  /* This will not be reached */
-            case SIGABRT:
-                sigdie(sig, "Unhandled SIGABRT: An abort() occurred.");
-                break;  /* This will not be reached */
-            case SIGFPE:
-                sigdie(sig, "Unhandled SIGFPE: An unhandled floating point exception occurred.");
-                break;  /* This will not be reached */
-            case SIGBUS:
-                sigdie(sig, "Unhandled SIGBUS: A bus error occurred.");
-                break;  /* This will not be reached */
-            case SIGSEGV:
-                sigdie(sig, "Unhandled SIGSEGV: A segmentation fault occurred.");
-                break;  /* This will not be reached */
-        };
-        sigdie(sig, "Unknown signal received.\n");
+        sigdie_for_sig(sig, inside);
     }
 }
 
 
-extern int sig_raise_exception(int sig, const char* msg);
+/* A trampoline to jump to after handling a signal.
+ *
+ * The jump to sig_on() uses cylongjmp(), which does not restore the
+ * signal context. This is done for efficiency, as cysetjmp() is
+ * significantly faster this way. But in order to get away from our alt
+ * stack after handling a signal, we need an additional siglongjmp()
+ * call to restore the signal context. This is the call from the signal
+ * handler to this trampoline function.
+ *
+ * Setting this up requires some trickery:
+ * (A) create a separate stack for this trampoline function
+ * (B) start a new thread using this stack
+ * (C) set a jump point on the trampoline stack using cysetjmp()
+ * (D) exit the thread
+ * (E) back in the main thread, jump to the point set at (C). Now we are
+ *     on the trampoline stack
+ * (F) set a jump point with savesigs=1. This is where we will jump to
+ *     after handling a signal
+ * (G) jump back to the main program
+ *
+ * NOTE: it may look strange to use threads for this, but there are not
+ * a lot of good ways to get code running on an arbitrary stack. In
+ * fact, POSIX recommends threads in
+ * http://pubs.opengroup.org/onlinepubs/009695299/functions/makecontext.html
+ */
+static void* _sig_on_trampoline(void* dummy)
+{
+    register int sig;
+
+    /* Reserve some unused stack space to prevent pthread_exit() from
+     * clobbering the stack that we care about. This is in particular
+     * needed on certain older GNU/Linux systems:
+     * https://trac.sagemath.org/ticket/25092#comment:6 */
+    char stack_guard[2048];
+
+    if (cysetjmp(trampoline_setup) == 0)
+        /* The argument to pthread_exit() does not matter. We use
+         * stack_guard to prevent GCC from optimizing away the
+         * stack_guard variable. */
+        pthread_exit(stack_guard);
+
+    sig = sigsetjmp(trampoline, 1);
+    reset_CPU();
+    cylongjmp(cysigs.env, sig);
+}
+
+
+static void setup_trampoline(void)
+{
+    int ret;
+    pthread_t child;
+    pthread_attr_t attr;
+    void* trampolinestack;
+    size_t trampolinestacksize = 1 << 16;
+
+    while (trampolinestacksize < PTHREAD_STACK_MIN) trampolinestacksize *= 2;
+    trampolinestack = mmap(NULL, trampolinestacksize,
+            PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+    if (trampolinestack == MAP_FAILED) {perror("mmap"); exit(1);}
+
+    ret = pthread_attr_init(&attr);
+    if (ret) {errno = ret; perror("pthread_attr_init"); exit(1);}
+    ret = pthread_attr_setstack(&attr, trampolinestack, trampolinestacksize);
+    if (ret) {errno = ret; perror("pthread_attr_setstack"); exit(1);}
+    ret = pthread_create(&child, &attr, _sig_on_trampoline, NULL);
+    if (ret) {errno = ret; perror("pthread_create"); exit(1);}
+    pthread_attr_destroy(&attr);
+    ret = pthread_join(child, NULL);
+    if (ret) {errno = ret; perror("pthread_join"); exit(1);}
+
+    if (cysetjmp(cysigs.env) == 0)
+    {
+        cylongjmp(trampoline_setup, 1);
+    }
+}
+
 
 /* This calls sig_raise_exception() to actually raise the exception. */
 static void do_raise_exception(int sig)
@@ -256,7 +370,7 @@ static void _sig_on_interrupt_received(void)
     sigprocmask(SIG_SETMASK, &oldset, NULL);
 }
 
-/* Cleanup after siglongjmp() (reset signal mask to the default, set
+/* Cleanup after cylongjmp() (reset signal mask to the default, set
  * sig_on_count to zero) */
 static void _sig_on_recover(void)
 {
@@ -286,39 +400,49 @@ static void _sig_off_warning(const char* file, int line)
 }
 
 
+static void setup_alt_stack(void)
+{
+    /* Static space for the alternate signal stack. The size should be
+     * of the form MINSIGSTKSZ + constant. The constant is chosen rather
+     * ad hoc but sufficiently large. */
+    static char alt_stack[MINSIGSTKSZ + 5120 + BACKTRACELEN * sizeof(void*)];
+
+    stack_t ss;
+    ss.ss_sp = alt_stack;
+    ss.ss_size = sizeof(alt_stack);
+    ss.ss_flags = 0;
+    if (sigaltstack(&ss, NULL) == -1) {perror("sigaltstack"); exit(1);}
+#if defined(__CYGWIN__) && defined(__x86_64__)
+    cygwin_setup_alt_stack();
+#endif
+}
+
+
 static void setup_cysignals_handlers(void)
 {
+    struct sigaction sa;
+    memset(&sa, 0, sizeof(sa));
+
     /* Reset the cysigs structure */
     memset(&cysigs, 0, sizeof(cysigs));
 
-    /* Save the default signal mask */
-    sigprocmask(SIG_BLOCK, NULL, &default_sigmask);
-
-    /* Save the signal mask with non-critical signals blocked */
-    sigprocmask(SIG_BLOCK, NULL, &sigmask_with_sigint);
-    sigaddset(&sigmask_with_sigint, SIGHUP);
-    sigaddset(&sigmask_with_sigint, SIGINT);
-    sigaddset(&sigmask_with_sigint, SIGALRM);
-
-    /* Allocate alternate signal stack */
-    stack_t ss;
-    ss.ss_size = 1 << 14;
-    ss.ss_sp = malloc(ss.ss_size);
-    if (ss.ss_sp == NULL) {perror("malloc"); exit(1);}
-    ss.ss_flags = 0;
-    if (sigaltstack(&ss, NULL) == -1) {perror("sigaltstack"); exit(1);}
-
-    /* Install signal handlers */
-    struct sigaction sa;
-    memset(&sa, 0, sizeof(sa));
-    /* Block non-critical signals during the signal handlers */
-    sigemptyset(&sa.sa_mask);
+    /* Block non-critical signals during the signal handlers and while
+     * cleaning up after handling a signal */
     sigaddset(&sa.sa_mask, SIGHUP);
     sigaddset(&sa.sa_mask, SIGINT);
     sigaddset(&sa.sa_mask, SIGALRM);
 
+    /* Save the default signal mask and apply the signal mask with
+     * non-critical signals now to save it on the trampoline.
+     * After setting up the trampoline, we reset the signal mask. */
+    sigprocmask(SIG_BLOCK, &sa.sa_mask, &default_sigmask);
+    setup_trampoline();
+    sigprocmask(SIG_SETMASK, &default_sigmask, &sigmask_with_sigint);
+
+    /* Install signal handlers */
     /* Handlers for interrupt-like signals */
     sa.sa_handler = cysigs_interrupt_handler;
+    sa.sa_flags = 0;
     if (sigaction(SIGHUP, &sa, NULL)) {perror("sigaction"); exit(1);}
     if (sigaction(SIGINT, &sa, NULL)) {perror("sigaction"); exit(1);}
     if (sigaction(SIGALRM, &sa, NULL)) {perror("sigaction"); exit(1);}
@@ -339,19 +463,22 @@ static void setup_cysignals_handlers(void)
 
 static void print_sep(void)
 {
-    fprintf(stderr,
-        "------------------------------------------------------------------------\n");
+    fputs("------------------------------------------------------------------------\n",
+            stderr);
     fflush(stderr);
 }
 
 /* Print a backtrace if supported by libc */
 static void print_backtrace()
 {
-    void* backtracebuffer[1024];
+    void* backtracebuffer[BACKTRACELEN];
     fflush(stderr);
 #if HAVE_BACKTRACE
-    int btsize = backtrace(backtracebuffer, 1024);
-    backtrace_symbols_fd(backtracebuffer, btsize, 2);
+    int btsize = backtrace(backtracebuffer, BACKTRACELEN);
+    if (btsize)
+        backtrace_symbols_fd(backtracebuffer, btsize, 2);
+    else
+        fputs("(no backtrace available)\n", stderr);
     print_sep();
 #endif
 }
@@ -421,8 +548,11 @@ static void sigdie(int sig, const char* s)
      * the user is probably using other debugging tools and we don't
      * want to interfere with that. */
 #else
-#ifndef __APPLE__
+#if !(defined(__APPLE__) || defined(__CYGWIN__))
     /* See http://trac.sagemath.org/13889 for how Apple screwed this up */
+    /* On Cygwin this has never quite worked, and in particular when run
+       from the altstack handler it just results in fork errors, so disable
+       this feature for now */
     if (getenv("CYSIGNALS_CRASH_NDEBUG") == NULL)
         print_enhanced_backtrace();
 #endif
@@ -438,7 +568,9 @@ static void sigdie(int sig, const char* s)
     }
 
 dienow:
-    /* Suicide with signal ``sig`` */
+    /* Suicide with signal ``sig``. We intentionally use kill(getpid())
+     * instead of raise() because we want to kill all threads in a
+     * multi-threaded program. */
     kill(getpid(), sig);
 
     /* We should be dead! */
